// ===========================================================================
// CONTENT  : INTERFACE ICheckedCharset
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 20/03/2016
// HISTORY  :
//  20/03/2016  mdu  CREATED
//
// Copyright (c) 2016, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.bif.text ;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.nio.charset.Charset;

/**
 * Specifies the interface of objects that represents a character encoding name 
 * that has already been checked if it is supported by the platform.
 * <p>
 * It helps to avoid the annoying try-catch code-pollution every time you
 * use a valid encoding, but the API requires the handling of UnsupportedEncodingException
 * even if it cannot occur at all.  
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public interface ICheckedCharset
{
  /**
   * Returns the pure upper case encoding name (e.g. "UTF-8").
   */
  public String getName();

  /**
   * Returns the corresponding charset object.
   */
  public Charset getCharset();

  /**
   * Returns a new String initialized with the given bytes which are expected 
   * to be encoded according to the character set this instance represents,
   * avoiding the unnecessary exception handling (UnsupportedEncodingException).
   * <p>
   * It is the same like: new String(bytes, checkedCharsetName.toString());
   * 
   * @param bytes The byte to create a string from.
   * @return A new string built from the given bytes that were encoded with the charset of this instance.
   */
  public String newString(byte[] bytes);

  /**
   * Returns the bytes of the given string as array encoded to the charset 
   * this instance represents, avoiding the unnecessary exception handling
   * (UnsupportedEncodingException).
   * 
   * @param string The string to be converted to a byte array using this instance's charset.
   * @return The encoded byte array.
   */
  public byte[] getBytes(String string);
}