// ===========================================================================
// CONTENT  : CLASS NamedOutputStream
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2015
// HISTORY  :
//  14/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Closeable;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;

/**
 * A container for an OutputStream with an associated name.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class NamedOutputStream extends NamedCloseable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final OutputStream outputStream;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Opens the provided file and returns the output stream along with the file name.
   * 
   * @param file The file to be opened (must not be null).
   * @return An object that contains the (normalized) absolute file name and the
   * FileOutputStream on the just opened file.
   * @throws FileNotFoundException If the provided file cannot be created/opened.
   */
  public static NamedOutputStream open(File file) throws FileNotFoundException 
  {
    FileOutputStream outStream;
    String name;
    
    outStream = new FileOutputStream(file);
    name = FileUtil.current().standardize(file.getAbsolutePath());
    return new NamedOutputStream(name, outStream);
  }
  
  // -------------------------------------------------------------------------

  /**
   * Opens the file specified by the given filename and returns the output stream along with the file name.
   * 
   * @param filename The file to be opened (must not be null).
   * @return An object that contains the (normalized) absolute file name and the
   * FileOutputStream on the just opened file.
   * @throws FileNotFoundException If a file with the given name cannot be created/opened.
   */
  public static NamedOutputStream open(String filename) throws FileNotFoundException 
  {
    File file;
    
    file = new File(filename);
    return open(file);
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Create a new instance with the given name associated to the given output stream.
   * 
   * @param name The name of the output stream (must not be null).
   * @param stream The output stream to be associated with the name (must not be null).
   * @throws IllegalArgumentException If either name or stream is null.
   */
  public NamedOutputStream(String name, OutputStream stream)
  {
    super(name);
    if (stream == null)
    {
      throw new IllegalArgumentException("The provided argument for stream is null.");
    }
    this.outputStream = stream;
  } 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the underlying OutputStream (will never be null).
   */
  public OutputStream getOutputStream()
  {
    return this.outputStream;
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  @Override
  protected Closeable getCloseable()
  {
    return this.getOutputStream();
  }
  
} // class NamedOutputStream
