// ===========================================================================
// CONTENT  : CLASS NamedReader
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 14/05/2015
// HISTORY  :
//  14/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.Charset;

/**
 * A container for a Reader with an associated name.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class NamedReader extends NamedCloseable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Reader reader;
  
  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Opens the provided file and returns the reader along with the file name.
   * 
   * @param file The file to be opened (must not be null).
   * @param charset The character encoding of the file.
   * @return An object that contains the (normalized) absolute file name and the
   * InputStreamReader on the just opened file.
   * @throws FileNotFoundException If the provided file cannot be found.
   */
  public static NamedReader open(File file, Charset charset) throws FileNotFoundException 
  {
    InputStreamReader streamReader;
    String name;
    
    streamReader = new InputStreamReader(new FileInputStream(file), charset);

    name = FileUtil.current().standardize(file.getAbsolutePath());
    return new NamedReader(name, streamReader);
  }
  
  // -------------------------------------------------------------------------

  /**
   * Opens the file specified by the given filename and returns the reader along with the file name.
   * 
   * @param filename The file to be opened (must not be null).
   * @param charset The character encoding of the file.
   * @return An object that contains the (normalized) absolute file name and the
   * InputStreamReader on the just opened file.
   * @throws FileNotFoundException If a file with the given name cannot be found.
   */
  public static NamedReader open(String filename, Charset charset) throws FileNotFoundException 
  {
    File file;
    
    file = new File(filename);
    return open(file, charset);
  }
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Create a new instance with the given name associated to the given reader.
   * 
   * @param name The name of the reader (must not be null).
   * @param reader The reader to be associated with the name (must not be null).
   * @throws IllegalArgumentException If either name or reader is null.
   */
  public NamedReader(String name, Reader reader)
  {
    super(name);
    if (reader == null)
    {
      throw new IllegalArgumentException("The provided argument for reader is null.");
    }
    this.reader = reader;
  } 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the underlying Reader (will never be null).
   */
  public Reader getReader()
  {
    return this.reader;
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  @Override
  protected Closeable getCloseable()
  {
    return null;
  }
  
} // class NamedReader
