// ===========================================================================
// CONTENT  : ABSTRACT CLASS AbstractContainer
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 01/11/2015
// HISTORY  :
//  01/11/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file.cfiles;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.util.ArrayList;
import java.util.List;

/**
 * Abstract implementation of a container that holds multiple items.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
abstract public class AbstractContainer implements IContainer
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String name;
  private final List<IItem> itemList = new ArrayList<IItem>();

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected AbstractContainer()
  {
    super();
  }

  protected AbstractContainer(String name)
  {
    super();
    this.setName(name);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  @Override
  public List<IItem> getItems()
  {
    return new ArrayList<IItem>(this.getItemList());
  }

  @Override
  public int size()
  {
    return this.getItemList().size();
  }

  @Override
  public boolean isEmpty()
  {
    return this.getItemList().isEmpty();
  }

  @Override
  public void addItem(IItem... items)
  {
    if (items != null)
    {
      for (IItem item : items)
      {
        this.basicAdd(item);
      }
    }
  }

  @Override
  public void removeItem(IItem... items)
  {
    if (items != null)
    {
      for (IItem item : items)
      {
        this.basicRemove(item);
      }
    }
  }

  @Override
  public void release()
  {
    IItem[] items;

    items = new IItem[this.size()];
    this.getItemList().toArray(items);
    this.removeItem(items);
  }

  @Override
  public String getName()
  {
    return this.name;
  }
  
  @Override
  public void appendNameTo(StringBuilder builder)
  {
    if (this.getName() == null)
    {
      throw new IllegalStateException("Name currently not set. Builder contents: " + builder.toString());
    }
    builder.append(this.getName());
  }

  public void setName(String name)
  {
    this.name = name;
  }
  
  /**
   * This is a debug method and to intended to be used for business logic.
   */
  @Override
  public String toString()
  {
    return this.getClass().getSimpleName() + "(\"" + this.getName() + "\")";
  }
  
  @Override
  public int hashCode()
  {
    return this.basicHashCode();
  }
  
  @Override
  public boolean equals(Object obj)
  {
    return this.basicEquals(obj);
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Adds the given item if it s not null and if it is not yet in the item list
   * and sets this container as the container of the given item.
   * 
   * @param item The item to add.
   */
  protected void basicAdd(IItem item)
  {
    if (item != null)
    {
      if (!this.contains(item))
      {
        this.getItemList().add(item);
        item.setContainer(this);
      }
    }
  }

  /**
   * Adds the given item if it s not null and if it is not yet in the item list.
   * 
   * @param item The item to add.
   */
  protected void basicRemove(IItem item)
  {
    boolean removed;

    if (item != null)
    {
      removed = this.getItemList().remove(item);
      if (removed)
      {
        item.release();
      }
    }
  }

  protected boolean contains(IItem item)
  {
    if (item != null)
    {
      return this.getItemList().contains(item);
    }
    return false;
  }

  /**
   * Calculates just the hash code from the underlying name.
   */
  protected int basicHashCode()
  {
    if (this.getName() == null)
    {
      return 0;
    }
    return this.getName().hashCode();
  }
  
  /**
   * Compares just the names of the objects.
   */
  protected boolean basicEquals(Object obj)
  {
    IContainer other;

    if (obj instanceof IContainer)
    {
      other = (IContainer)obj;
      if ((this.getName() == null))
      {
        return (other.getName() == null);
      }
      return this.getName().equals(other.getName());
    }
    return false;
  }
  
  protected List<IItem> getItemList()
  {
    return this.itemList;
  }
}
