// ===========================================================================
// CONTENT  : CLASS MediaType
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 19/08/2014
// HISTORY  :
//  19/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net;

// ===========================================================================
// IMPORTS
// ===========================================================================
import org.pfsw.bif.text.IStringPair;
import org.pfsw.text.StringUtil;

/**
 * A definition of internet media types that are comprised of a type and a sub-type.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class MediaType
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final String MEDIA_TYPE_WILDCARD = "*";

  public final static MediaType WILDCARD_TYPE = new MediaType(MEDIA_TYPE_WILDCARD, MEDIA_TYPE_WILDCARD);
  public final static String WILDCARD = WILDCARD_TYPE.asString();

  public final static MediaType TEXT_PLAIN_TYPE = new MediaType("text", "plain");
  public final static String TEXT_PLAIN = TEXT_PLAIN_TYPE.asString();

  public final static MediaType TEXT_HTML_TYPE = new MediaType("text", "html");
  public final static String TEXT_HTML = TEXT_HTML_TYPE.asString();

  public final static MediaType APPLICATION_FORM_URLENCODED_TYPE = new MediaType("application", "x-www-form-urlencoded");
  public final static String APPLICATION_FORM_URLENCODED = APPLICATION_FORM_URLENCODED_TYPE.asString();

  public final static MediaType MULTIPART_FORM_DATA_TYPE = new MediaType("multipart", "form-data");
  public final static String MULTIPART_FORM_DATA = MULTIPART_FORM_DATA_TYPE.asString();

  public final static MediaType APPLICATION_ATOM_XML_TYPE = new MediaType("application", "atom+xml");
  public final static String APPLICATION_ATOM_XML = APPLICATION_ATOM_XML_TYPE.asString();

  public final static MediaType APPLICATION_JSON_TYPE = new MediaType("application", "json");
  public final static String APPLICATION_JSON = APPLICATION_JSON_TYPE.asString();

  public final static MediaType APPLICATION_XML_TYPE = new MediaType("application", "xml");
  public final static String APPLICATION_XML = APPLICATION_XML_TYPE.asString();

  public final static MediaType APPLICATION_XHTML_XML_TYPE = new MediaType("application", "xhtml+xml");
  public final static String APPLICATION_XHTML_XML = APPLICATION_XHTML_XML_TYPE.asString();

  public final static MediaType APPLICATION_OCTET_STREAM_TYPE = new MediaType("application", "octet-stream");
  public final static String APPLICATION_OCTET_STREAM = APPLICATION_OCTET_STREAM_TYPE.asString();

  private static final String SEPARATOR = "/";
  
  private static final MediaType[] ALL = { WILDCARD_TYPE, TEXT_PLAIN_TYPE, TEXT_HTML_TYPE, 
    APPLICATION_FORM_URLENCODED_TYPE, MULTIPART_FORM_DATA_TYPE, APPLICATION_ATOM_XML_TYPE, 
    APPLICATION_JSON_TYPE, APPLICATION_XML_TYPE,
    APPLICATION_XHTML_XML_TYPE, APPLICATION_OCTET_STREAM_TYPE 
  };

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private String type;
  private String subType;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the MediType object for the given type string. 
   * 
   * @param typeDef The type definition (e.g. "image/jpg") (must not be null).
   * @throws IllegalArgumentException If the given typeDef is null or otherwise no valid media type definition.
   */
  public static MediaType get(String typeDef)
  {
    for (MediaType mediaType : ALL)
    {
      if (mediaType.isType(typeDef))
      {
        return mediaType;
      }
    }
    return new MediaType(typeDef);
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Creates a new MediaType.
   * 
   * @param type The main type (must not be null).
   * @param subType The sub type (must not be null).
   * @throws IllegalArgumentException If any of the arguments is null.
   */
  public MediaType(String type, String subType)
  {
    super();
    this.init(type, subType);
  }

  /**
   * Creates a new MediaType.
   * 
   * @param typeDef The type definition comprised of a main type and sub type separated by a slash (must not be null).
   * @throws IllegalArgumentException If the arguments is null or blank or no valid media type definition.
   */
  public MediaType(String typeDef)
  {
    super();
    if (StringUtil.current().isNullOrBlank(typeDef))
    {
      throw new IllegalArgumentException("Argument typeDef is null or blank!");
    }
    IStringPair pair = StringUtil.current().splitStringPair(typeDef, SEPARATOR);
    this.init(pair.getString1(), pair.getString2());
  } 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the string representation (e.g. "text/html") of this media type.
   */
  public String asString()
  {
    return this.getType() + SEPARATOR + this.getSubType();
  }

  /**
   * Returns true if the given string equals (case-insensitive) the receiver's
   * string representation.
   */
  public boolean isType(String typeDef)
  {
    return this.asString().equalsIgnoreCase(typeDef);
  }

  @Override
  public String toString()
  {
    return "MediaType(" + this.asString() + ")";
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == null)
    {
      return false;
    }
    if (!(obj instanceof MediaType))
    {
      return false;
    }
    MediaType other = (MediaType)obj;
    return (this.getType().equalsIgnoreCase(other.getType()) && this.getSubType().equalsIgnoreCase(other.getSubType()));
  }

  @Override
  public int hashCode()
  {
    return this.asString().toLowerCase().hashCode();
  }

  public boolean isWildcardType()
  {
    return this.getType().equals(MEDIA_TYPE_WILDCARD);
  }

  public boolean isWildcardSubType()
  {
    return this.getSubType().equals(MEDIA_TYPE_WILDCARD);
  }

  public String getType()
  {
    return type;
  }

  public String getSubType()
  {
    return subType;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void init(String main, String sub)
  {
    if (StringUtil.current().isNullOrBlank(main))
    {
      throw new IllegalArgumentException("Specified type is null or blank!");
    }
    if (StringUtil.current().isNullOrBlank(sub))
    {
      throw new IllegalArgumentException("Specified subType is null or blank!");
    }
    this.setType(main.trim());
    this.setSubType(sub.trim());
  }

  protected void setType(String type)
  {
    this.type = type;
  }

  protected void setSubType(String subType)
  {
    this.subType = subType;
  }

} // class MediaType
