// ===========================================================================
// CONTENT  : CLASS NetUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 10/05/2015
// HISTORY  :
//  20/08/2014  mdu  CREATED
//  10/05/2015  mdu  added  --> withLeadingSlash(), withTrailingSlash(), withoutLeadingSlash(), withoutTrailingSlash()
//
// Copyright (c) 2014-2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net ;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Closeable;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;

import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * Utility and convenience methods for network programming, provided as singleton.
 * The class has a protected default constructor that allows to extend it and
 * potentially override methods.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public class NetUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected static final char SLASH = '/';

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static NetUtil soleInstance = new NetUtil();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static NetUtil current()
  {
    return soleInstance ;
  } // current() 

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected NetUtil()
  {
    super() ;
  } // NetUtil() 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Silently closes the given object. 
   * If the given object is null, nothing will be done. If an IOException occurs
   * false will be returned.
   * 
   * @param closeable The object to close (may be null).
   * @return true if executed without problem, false if an IOException occurred.
   */
  public boolean close(Closeable closeable) 
  {
    if (closeable != null)
    {
      try
      {
        closeable.close();
      }
      catch (IOException ex)
      {
        return false;
      }
    }
    return true;
  } // close() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the given string character and URL encoded.
   * 
   * @param string The string to be encoded (if null, null will be returned).
   * @param charset The character encoding to be used on the string before doing the URL encoding (must not be null).
   * @return The encoded string or null in case of a completely unexpected UnsupportedEncodingException.
   */
  public String urlEncode(String string, Charset charset) 
  {
    if (string == null)
    {
      return null;
    }
    try
    {
      return URLEncoder.encode(string, charset.name());
    }
    catch (UnsupportedEncodingException ex)
    {
      // Should never happen because charset is already checked for supported encoding!
      LoggerProvider.getComponentLogger().logError("[" + LoggerProvider.PFNET0003 
          + "] org.pfsw.net.NetUtil.urlEncode(String, Charset): ", ex);
      return null;
    }
  } // urlEncode() 
  
  // -------------------------------------------------------------------------

  /**
   * Returns the given string character and URL decoded.
   * 
   * @param string The string to be decoded (if null, null will be returned).
   * @param charset The character encoding to be used on the string after the URL decoding (must not be null).
   * @return The decoded string or null in case of a completely unexpected UnsupportedEncodingException.
   */
  public String urlDecode(String string, Charset charset) 
  {
    if (string == null)
    {
      return null;
    }
    try
    {
      return URLDecoder.decode(string, charset.name());
    }
    catch (UnsupportedEncodingException ex)
    {
      // Should never happen because charset is already checked for supported encoding!
      LoggerProvider.getComponentLogger().logError("[" + LoggerProvider.PFNET0004
          + "] org.pfsw.net.NetUtil.urlDecode(String, Charset): ", ex);
      return null;
    }
  } // urlDecode() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the given string guaranteed with a leading slash. 
   * @param string The string that should have a leading slash (must not be null).
   */
  public String withLeadingSlash(String string) 
  {
    return this.str().withLeadingChar(string, SLASH);
  } // withLeadingSlash()
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the given string guaranteed with a trailing slash. 
   * @param string The string that should have a trailing slash (must not be null).
   */
  public String withTrailingSlash(String string) 
  {
    return this.str().withTrailingChar(string, SLASH);
  } // withTrailingSlash()
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the given string guaranteed with no leading slash. 
   * @param string The string that should not have a leading slash (must not be null).
   */
  public String withoutLeadingSlash(String string) 
  {
    return this.str().withoutLeadingChar(string, SLASH);
  } // withoutLeadingSlash()
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the given string guaranteed with no trailing slash. 
   * @param string The string that should not have a trailing slash (must not be null).
   */
  public String withoutTrailingSlash(String string) 
  {
    return this.str().withoutTrailingChar(string, SLASH);
  } // withoutTrailingSlash()
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected StringUtil str()
  {
    return StringUtil.current();
  } // str()

  //-------------------------------------------------------------------------
  
} // class NetUtil 
