// ===========================================================================
// CONTENT  : CLASS ClientRequest
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 19/08/2014
// HISTORY  :
//  19/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.net.URL;
import java.util.Map;

import org.pfsw.net.NetUtil;

/**
 * Definition of a client request that can be executed with a UserAgent.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class ClientRequest extends CommunicationData
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final NetUtil NU = NetUtil.current();
  
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private HttpMethod httpMethod = HttpMethod.GET;
  private URL url;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public ClientRequest()
  {
    super();
  } // ClientRequest()

  // -------------------------------------------------------------------------
  
  public ClientRequest(URL url)
  {
    this(HttpMethod.GET, url);
  } // ClientRequest()
  
  // -------------------------------------------------------------------------
  
  public ClientRequest(HttpMethod method, URL url)
  {
    this();
    this.setURL(url);
    this.setHttpMethod(method);
  } // ClientRequest()
  
  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public HttpMethod getHttpMethod()
  {
    return httpMethod;
  }
  
  // -------------------------------------------------------------------------

  public void setHttpMethod(HttpMethod httpMethod)
  {
    this.httpMethod = httpMethod;
  }
  
  // -------------------------------------------------------------------------

  public URL getURL()
  {
    return this.url;
  }
  
  // -------------------------------------------------------------------------

  public void setURL(URL url)
  {
    this.url = url;
  }
  
  // -------------------------------------------------------------------------

  @Override
  public String toString()
  {
    return "ClientRequest(" + this.getURL() + ")";
  }

  // -------------------------------------------------------------------------
  
  /**
   * Adds the given media types (e.g. "text/html") to the ACCEPT request header.
   */
  public void accept(String... mediaTypes)
  {
    for (String mediaType : mediaTypes)
    {
      this.addHeader(RequestHeaderNames.ACCEPT, mediaType);
    }
  }
  
  // -------------------------------------------------------------------------
  
  /**
   * Appends the given map as content key/value pairs. All keys and values
   * will automatically be URL encoded using the currently defined character encoding. 
   */
  public void appendContent(Map<String, String> parameters) 
  {
    this.checkIfHttpMethodAllowsBody();
    for (Map.Entry<String, String> param : parameters.entrySet())
    {
      if (this.getContentLength() > 0)
      {
        this.appendContent("&");
      }
      this.appendContent(NU.urlEncode(param.getKey(), this.getCharsetEncoding()));
      this.appendContent("=");
      this.appendContent(NU.urlEncode(param.getValue(), this.getCharsetEncoding()));
    }
  }
  
  // -------------------------------------------------------------------------

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void checkIfHttpMethodAllowsBody() 
  {
    if (!this.getHttpMethod().hasBody())
    {
      throw new IllegalStateException("The current request method " + this.getHttpMethod().getName() 
          + " does not allow content!");
    }
  }
  
  // -------------------------------------------------------------------------

} // class ClientRequest
