// ===========================================================================
// CONTENT  : ABSTRACT CLASS CommunicationData
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 30/12/2016
// HISTORY  :
//  20/08/2014  mdu  CREATED
//  301/12/2016 mdu   changed -> all around content type and charset encoding
//
// Copyright (c) 2014-2016, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.List;
import java.util.Map;

import org.pfsw.bif.text.ICharsets;
import org.pfsw.bif.text.IStringPair;
import org.pfsw.logging.Logger;
import org.pfsw.net.MediaType;
import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * Common fields and methods for request and response objects in HTTP communication.
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
abstract class CommunicationData
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();
  private static final String CONTENT_TYPE = RequestHeaderNames.CONTENT_TYPE;
  private static final String CHARSET_SEPARATOR = ";";
  private static final String CHARSET_PARAMETER = "charset";
  private static final Charset DEFAULT_CHARSET = ICharsets.UTF_8;
  
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final Headers headers = new Headers();
  private ByteArrayOutputStream contentDataStream = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected CommunicationData()
  {
    super();
  } // CommunicationData()

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Sets the "Content-Type" header.
   * If mediaType is null the header will be removed.
   * @param mediaType The media type to set. If null the header will be removed.
   * @param charset The charset to be set. If null it will be ignored.
   */
  public void setContentType(MediaType mediaType, Charset charset)
  {
    this.setContentTypeHeader(mediaType, charset);
  }
  
  /**
   * Returns the character encoding defined in the Content-Type header
   * or null if it is not set.
   */
  public Charset getCharset()
  {
    return this.asCharset(this.getContentTypeHeader());
  }
  
  /**
   * Sets the character encoding in the Content-Type header
   * or removes it if the given parameter is null.
   * @throws IllegalStateException If no media type is specified yet.
   */
  public void setCharset(Charset charset)
  {
    MediaType mediaType;
    
    mediaType = this.getContentMediaType(); // Preserve media type definition
    if ((mediaType == null) && (charset != null))
    {
      throw new IllegalStateException("Setting a charset without a media type is not allowed!");
    }
    this.setContentTypeHeader(mediaType, charset);
  }

  /**
   * Returns the media type specified in the "Content-Type" header
   * of null if not set. 
   */
  public MediaType getContentMediaType()
  {
    return this.asMediaType(this.getContentTypeHeader());
  }
  
  /**
   * Sets the media type in the Content-Type header
   * or removes the entire header if the given parameter is null.
   */
  public void setContentMediaType(MediaType mediaType)
  {
    Charset charset;
    
    charset = this.getCharset(); // Preserve charset definition
    this.setContentTypeHeader(mediaType, charset);
  }

  /**
   * Returns the character encoding defined in the Content-Type header
   * or null if it is not set.
   * @deprecated Use {@link #getCharset()} instead.
   */
  @Deprecated
  public Charset getEncoding()
  {
    return this.getCharset();
  }

  /**
   * Sets the character encoding in the Content-Type header
   * or removes it if the given parameter is null.
   * @deprecated Use {@link #setCharset(Charset)} instead.
   */
  @Deprecated
  public void setEncoding(Charset charset)
  {
    this.setCharset(charset);
  }

  /**
   * Returns the media type specified in the "Content-Type" header
   * of null if not set. 
   * 
   * @deprecated Use {@link #getContentMediaType()} instead.
   */
  @Deprecated
  public MediaType getContentType()
  {
    return this.getContentMediaType();
  }
  
  /**
   * Sets the media type in the Content-Type header
   * or removes the entire header if the given parameter is null.
   * @deprecated Use {@link #setContentMediaType(MediaType)} instead.
   */
  @Deprecated
  public void setContentType(MediaType mediaType)
  {
    this.setContentMediaType(mediaType);
  }

  // -------------------------------------------------------------------------
  
  public boolean hasContentData()
  {
    return (this.getContentDataStream() != null) && (this.getContentDataStream().size() > 0);
  }

  // -------------------------------------------------------------------------

  public int getContentLength()
  {
    if (this.hasContentData())
    {
      return this.getContentBuffer().size();
    }
    return 0;
  }

  // -------------------------------------------------------------------------

  /**
   * Appends the given bytes to the content. 
   */
  public void appendContent(byte[] data)
  {
    if ((data != null) && (data.length > 0))
    {
      this.getContentBuffer().write(data, 0, data.length);
    }
  }

  // -------------------------------------------------------------------------

  /**
   * Appends the bytes from the given stream to the content. 
   * 
   * @param inStream The stream from which to read the bytes (must not be null).
   * @throws IOException If any problem with the input stream occurs.
   */
  public void appendContent(InputStream inStream) throws IOException
  {
    byte[] buffer;
    int numBytes;

    buffer = new byte[this.getReadBufferSize()];
    numBytes = inStream.read(buffer);
    while (numBytes != -1)
    {
      this.getContentBuffer().write(buffer, 0, numBytes);
      numBytes = inStream.read(buffer);
    }
  }

  // -------------------------------------------------------------------------

  /**
   * Appends the given text to the content using the currently defined
   * character encoding.
   * 
   * @param text The text to be appended (null will be ignored).
   * @throws IllegalStateException If no encoding is set.
   */
  public void appendContent(String text)
  {
    byte[] data;

    if (text == null)
    {
      return;
    }
    if (this.getCharsetEncoding() == null)
    {
      throw new IllegalStateException("Content cannot be set as string because no character encoding is defined yet!");
    }
    data = text.getBytes(this.getCharsetEncoding());
    this.appendContent(data);
  }

  // -------------------------------------------------------------------------

  /**
   * Returns the content data or null if no content is available.
   */
  public byte[] getContentData()
  {
    if (this.hasContentData())
    {
      return this.getContentBuffer().toByteArray();
    }
    return null;
  }

  // -------------------------------------------------------------------------

  /**
   * Returns the content data as reader (with correct char encoding) or null 
   * if no content is available.
   */
  public Reader getContentReader()
  {
    ByteArrayInputStream inStream;

    if (!this.hasContentData())
    {
      return null;
    }
    inStream = new ByteArrayInputStream(this.getContentData());
    return new BufferedReader(new InputStreamReader(inStream, this.getCharsetEncoding()));
  }

  // -------------------------------------------------------------------------

  /**
   * Returns the content data as text (with correct char encoding) or null 
   * if no content is available.
   */
  public String getContentText()
  {
    Reader reader;
    char[] buffer;
    int numChars;
    Writer result;

    if (!this.hasContentData())
    {
      return null;
    }
    buffer = new char[this.getReadBufferSize()];
    result = new StringWriter(this.getContentLength());
    reader = this.getContentReader();
    try
    {
      numChars = reader.read(buffer);
      while (numChars != -1)
      {
        result.write(buffer, 0, numChars);
        numChars = reader.read(buffer);
      }
    }
    catch (IOException ex)
    {
      this.getLogger().logError("[" + LoggerProvider.PFNET0002 + "] Failed to convert content to string.", ex);
    }
    return result.toString();
  }

  // -------------------------------------------------------------------------

  /**
   * Add all name/values from the given map.
   * 
   * @param headerMap The header names and values to be a added (if null, nothing will be done).
   */
  public void addAllHeaders(Map<String, List<String>> headerMap)
  {
    this.getHeaders().addAllHeaders(headerMap);
  }

  // -------------------------------------------------------------------------

  /**
   * Add a header value to the header with the given name. If the header does
   * not exist yet it will be created.
   */
  public void addHeader(String name, String value)
  {
    this.getHeaders().addHeader(name, value);
  } // addHeader() 

  // -------------------------------------------------------------------------

  /**
   * A a header with a single value. If the header does not exist yet it will 
   * be created, otherwise its value(s) will be replaced by the given value.
   */
  public void setHeader(String name, String value)
  {
    this.getHeaders().setHeader(name, value);
  } // setHeader() 

  // -------------------------------------------------------------------------

  /**
   * Returns a single value (the first) of the header with the given name. 
   * If the header does not exist null will be returned.
   */
  public String getHeader(String name)
  {
    return this.getHeaders().getHeader(name);
  } // getHeader() 

  // -------------------------------------------------------------------------

  /**
   * Returns the header with the given name and all its associated values. 
   */
  public void removeHeader(String name)
  {
    this.getHeaders().removeHeader(name);
  } // removeHeader() 

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected MediaType asMediaType(String contentType) 
  {
    IStringPair stringPair;
    
    if (SU.isNullOrBlank(contentType))
    {
      return null;
    }
    stringPair = SU.splitStringPair(contentType, CHARSET_SEPARATOR);
    if (SU.isNullOrBlank(stringPair.getString1()))
    {
      return null;
    }
    return MediaType.get(stringPair.getString1());
  }
  
  protected Charset asCharset(String contentType) 
  {
    IStringPair stringPair;

    if (SU.isNullOrBlank(contentType))
    {
      return null;
    }
    stringPair = SU.splitStringPair(contentType, CHARSET_SEPARATOR);
    if (SU.isNullOrBlank(stringPair.getString2()))
    {
      return null;
    }
    stringPair = SU.splitStringPair(stringPair.getString2(), "=");
    if ((!CHARSET_PARAMETER.equalsIgnoreCase(stringPair.getString1())) && SU.isNullOrBlank(stringPair.getString2()))
    {
      return null;
    }
    return Charset.forName(stringPair.getString2());
  }
  
  /**
   * Returns the "Content-Type" header unparsed.
   * That is with media type and charset.
   */
  protected String getContentTypeHeader() 
  {
    return this.getHeader(CONTENT_TYPE);
  }
  
  /**
   * Sets the "Content-Type" header.
   * If mediaType is null the header will be removed.
   */
  protected void setContentTypeHeader(MediaType mediaType, Charset charset)
  {
    String contentType;
    
    if (mediaType == null)
    {
      this.removeHeader(CONTENT_TYPE);
    }
    else
    {
      contentType = mediaType.asString();
      if (charset != null)
      {
        contentType = contentType + CHARSET_SEPARATOR + CHARSET_PARAMETER + "=" + charset.toString().toLowerCase();
      }
      this.setHeader(CONTENT_TYPE, contentType);
    }
  }
  
  protected Charset getCharsetEncoding() 
  {
    Charset charset;
    
    charset = this.getCharset();
    return (charset == null) ? this.getDefaultCharset() : charset;
  }
  
  protected Charset getDefaultCharset() 
  {
    return DEFAULT_CHARSET;
  }

  protected ByteArrayOutputStream getContentBuffer()
  {
    if (this.getContentDataStream() == null)
    {
      this.setContentDataStream(new ByteArrayOutputStream());
    }
    return this.getContentDataStream();
  }

  protected ByteArrayOutputStream getContentDataStream()
  {
    return this.contentDataStream;
  }

  protected void setContentDataStream(ByteArrayOutputStream bodyData)
  {
    this.contentDataStream = bodyData;
  }

  protected Headers getHeaders()
  {
    return headers;
  }

  protected int getReadBufferSize()
  {
    return 4096;
  }

  protected Logger getLogger()
  {
    return LoggerProvider.getComponentLogger();
  }

}
