// ===========================================================================
// CONTENT  : CLASS Headers
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 07/08/2014
// HISTORY  :
//  07/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.pfsw.util.DateUtil;

/**
 * Container for request and response headers.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class Headers extends CaseInsensitiveMultiValueMap
{
  private static final DateUtil DU = DateUtil.current();
  
  private static final int NOT_DEFINED_DATE = -1;
  private static final int NOT_DEFINED_INT = -1;

  public Headers()
  {
    super();
  }

  /**
   * Adds the given value to the values of the header with the specified name.
   * @param name Name of the header
   * @param value Value to be added
   */
  public void addHeader(String name, String value)
  {
    this.addKeyValue(name, value);
  }

  /**
   * Sets the given value as sole value of the header with
   * the specified name.
   * @param name The name of the header
   * @param value The single value to replace all existing values
   */
  public void setHeader(String name, String value)
  {
    this.setKeyValue(name, value);
  }
  
  /**
   * Add all name/values from the given map.
   * 
   * @param headers The header names and values to be a added (if null, nothing will be done).
   */
  public void addAllHeaders(Map<String, List<String>> headers) 
  {
    this.addAllFrom(headers);
  }

  public String getHeader(String name)
  {
    return this.getFirstValue(name);
  }

  public long getDateHeader(String name)
  {
    String value;
    
    value = this.getHeader(name);
    if (value == null)
    {
      return NOT_DEFINED_DATE;
    }
    return DU.parseDateRFC1123(value).getTime();
  }

  public int getIntHeader(String name)
  {
    String value = this.getHeader(name);
    if (value == null)
    {
      // Returning -1 is stupid but complies with the specification
      return NOT_DEFINED_INT;
    }
    return Integer.valueOf(value);
  }

  public void removeHeader(String name)
  {
    this.removeKey(name);
  }

  public String[] getHeaderNames()
  {
    return SU.asStrings(this.getKeys());
  }
  
  public Collection<String> getHeaderNamesCollection()
  {
    return this.getKeys();
  }

  @SuppressWarnings("unchecked")
  public String[] getHeaderValues(String name)
  {
    return SU.asStrings(this.getDataValues(name));
  }

  public Collection<String> getHeadersCollection(String name)
  {
    return this.getDataValues(name);
  }

  @Override
  public int hashCode()
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.getData().hashCode();
    return result;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
    {
      return true;
    }
    if (obj == null)
    {
      return false;
    }
    if (getClass() != obj.getClass())
    {
      return false;
    }
    Headers other = (Headers)obj;
    if (!this.getData().equals(other.getData()))
    {
      return false;
    }
    return true;
  }
  
  @Override
  public String toString()
  {
    return "Headers(" + this.getData() + ")";
  }

} // class Headers
