// ===========================================================================
// CONTENT  : CLASS HttpMethod
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 19/08/2014
// HISTORY  :
//  19/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Serializable;

/**
 * Represents an HTTP method like GET, PUT and so on.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class HttpMethod implements Serializable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 7440685816188511252L;

  public static final HttpMethod GET = new HttpMethod("GET", false);
  public static final HttpMethod POST = new HttpMethod("POST", true);
  public static final HttpMethod PUT = new HttpMethod("PUT", true);
  public static final HttpMethod DELETE = new HttpMethod("DELETE", false);
  public static final HttpMethod HEAD = new HttpMethod("HEAD", false);
  public static final HttpMethod OPTIONS = new HttpMethod("OPTIONS", false);
  public static final HttpMethod CONNECT = new HttpMethod("CONNECT", false);
  public static final HttpMethod TRACE = new HttpMethod("TRACE", false);
  public static final HttpMethod PATCH = new HttpMethod("PATCH", true); // See RFC-5789

  private static final HttpMethod[] HTTP11_METHODS = { GET, POST, PUT, DELETE, HEAD, OPTIONS, CONNECT, TRACE, PATCH };

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final String name;
  private final boolean hasBody;

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the corresponding HTTP/1.1 method that corresponds to the given name
   * (case-insensitive) or null if the name does not match any HTTP/1.1 method.
   */
  public static HttpMethod getHttp11Method(String methodName) 
  {
    for (HttpMethod httpMethod : HTTP11_METHODS)
    {
      if (httpMethod.isMethod(methodName))
      {
        return httpMethod;
      }
    }
    return null;
  }

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public HttpMethod(String name, boolean hasBody)
  {
    super();
    if ((name == null) || (name.trim().length() == 0))
    {
      throw new IllegalArgumentException("Required argument 'name' is not valid!");
    }
    this.name = name.trim();
    this.hasBody = hasBody;
  } // HttpMethod()

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public String getName()
  {
    return this.name;
  }

  public boolean hasBody()
  {
    return this.hasBody;
  }

  /**
   * Returns true if the given string equals (case-insensitive) this
   * HttpMethod's name.
   */
  public boolean isMethod(String method)
  {
    return this.getName().equalsIgnoreCase(method);
  }

  /**
   * Comparison is case-insensitive!
   */
  @Override
  public boolean equals(Object obj)
  {
    if (!(obj instanceof HttpMethod))
    {
      return false;
    }
    HttpMethod other = (HttpMethod)obj;
    return this.getName().equalsIgnoreCase(other.getName());
  }

  @Override
  public int hashCode()
  {
    return this.getName().toUpperCase().hashCode();
  }

  @Override
  public String toString()
  {
    return "HttpMethod(" + this.getName() + ")";
  }
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

} // class HttpMethod
