// ===========================================================================
// CONTENT  : CLASS ServerResponse
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 20/08/2014
// HISTORY  :
//  20/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.net.HttpURLConnection;
import java.nio.charset.Charset;
import java.nio.charset.UnsupportedCharsetException;

import org.pfsw.net.internal.LoggerProvider;

/**
 * Holds all data received from a server as response to a single HTTP request.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class ServerResponse extends CommunicationData
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final int responseCode;
  private String responseMessage;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public ServerResponse(int responseCode)
  {
    super();
    this.responseCode = responseCode;
  } // ServerResponse()

  // -------------------------------------------------------------------------
  
  public ServerResponse(int responseCode, String responseMessage)
  {
    this(responseCode);
    this.setResponseMessage(responseMessage);
  } // ServerResponse()
  
  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public int getResponseCode()
  {
    return responseCode;
  }
  
  // -------------------------------------------------------------------------

  public String getResponseMessage()
  {
    return responseMessage;
  }

  // -------------------------------------------------------------------------
  
  public void setResponseMessage(String responseMessage)
  {
    this.responseMessage = responseMessage;
  }
  
  // -------------------------------------------------------------------------
  
  /**
   * Sets the character encoding to be used on the received content when it gets 
   * converted to text.
   *  
   * @param charEncoding The encoding name (e.g. "UTF-8"). If null the default HTTP encoding will be set.
   */
  public void setCharEncoding(String charEncoding)
  {
    Charset enc;
    
    if (charEncoding == null)
    {
      enc = HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG;
    }
    else
    {
      try
      {
        enc = Charset.forName(charEncoding);
      }
      catch (UnsupportedCharsetException ex)
      {
        this.getLogger().logInfo("[{0}] Received unsupported character encoding {1}. Using default {2} instead.", 
            LoggerProvider.PFNET0001, charEncoding, HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG);
        enc = HttpUtil.HTTP_DEFAULT_CHAR_ENCODUNG;
      } 
    }      
    this.setCharset(enc);
  }
  
  // -------------------------------------------------------------------------

  /**
   * Returns true if the response has returned with the OK status code (i.e. 200).
   */
  public boolean isOk()
  {
    return this.isCodeIn(HttpURLConnection.HTTP_OK);
  }
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns true if the response has a status code that equals
   * one of the given codes.
   */
  public boolean isCodeIn(int... codes)
  {
    for (int code : codes)
    {
      if (this.getResponseCode() == code)
      {
        return true;
      }
    }
    return false;
  }
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns true if the response has a status code that is in the specified
   * code range (fromCode and toCode both included).
   */
  public boolean isCodeInRange(int fromCode, int toCode)
  {
    if ((this.getResponseCode() >= fromCode) && (this.getResponseCode() <= toCode))
    {
      return true;
    }
    return false;
  }
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

} // class ServerResponse
