// ===========================================================================
// CONTENT  : CLASS UserAgent
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.2 - 10.04.2016
// HISTORY  :
//  19/08/2014  mdu  CREATED
//  28/02/2016  mdu   changed --> Support chunked encoding
//  10/04/2016  mdu   added   --> Proxy support
//
// Copyright (c) 2014-2016, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.net.http;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.Proxy;
import java.net.URL;

import org.pfsw.logging.Logger;
import org.pfsw.net.internal.LoggerProvider;
import org.pfsw.text.StringUtil;

/**
 * A class that handles communication via HTTP for a client. That is, it
 * executes requests and receives the corresponding responses.
 *
 * @author Manfred Duchrow
 * @version 1.2
 */
public class UserAgent
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  // TODO Implement a CookieManager and set it here to a variable

  private final Headers headers = new Headers();
  private int readTimeout = 10000; // 10 seconds
  private int connectTimeout = 5000; // 5 seconds
  private String userAgentName = "PF-Net/1.0";
  private Proxy proxy = null;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public UserAgent()
  {
    this(null);
  }

  public UserAgent(Proxy proxy)
  {
    super();
    this.setProxy(proxy);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns the user agent's proxy. If null is provided, it means that
   * the system property settings regarding HTTP proxy are respected.
   */
  public Proxy getProxy()
  {
    return this.proxy;
  }

  /**
   * Sets the user agent's proxy. If null is provided, it means that
   * the system property settings regarding HTTP proxy are respected.
   */
  public void setProxy(Proxy proxy)
  {
    this.proxy = proxy;
  }

  /**
   * Add a header value to the header with the given name. If the header does
   * not exist yet it will be created.
   */
  public void addHeader(String name, String value)
  {
    this.getHeaders().addHeader(name, value);
  } // addHeader() 

  // -------------------------------------------------------------------------

  /**
   * A a header with a single value. If the header does not exist yet it will 
   * be created, otherwise its value(s) will be replaced by the given value.
   */
  public void setHeader(String name, String value)
  {
    this.getHeaders().setHeader(name, value);
  } // setHeader() 

  // -------------------------------------------------------------------------

  /**
   * Returns a single value (the first) of the header with the given name. 
   * If the header does not exist null will be returned.
   */
  public String getHeader(String name)
  {
    return this.getHeaders().getHeader(name);
  } // getHeader() 

  // -------------------------------------------------------------------------

  /**
   * Returns the header with the given name and all its associated values. 
   */
  public void removeHeader(String name)
  {
    this.getHeaders().removeHeader(name);
  } // removeHeader() 

  // -------------------------------------------------------------------------

  /**
   * Sends a request based on the given request definition and the user agent's
   * own state.
   * 
   * @param request The definition of the request to be sent (must not be null).
   * @return A response object containing all data received from the server.
   * @throws IOException Any problem during remote communication.
   */
  public ServerResponse sendRequest(ClientRequest request) throws IOException
  {
    return this.sendRequest(request, null);
  }

  /**
   * Sends a request based on the given request definition and the user agent's
   * own state.
   * 
   * @param request The definition of the request to be sent (must not be null).
   * @param responseStreamHandler An optional (may be null) handler for the response stream.
   * If provided the ServerResponse methods hasContentData(), getContentData(), getConentText(), 
   * getContentReader() must not be used anymore!
   * @return A response object containing the data received from the server except maybe the
   * content body if that was handled already by the given responseStreamHandler.
   * @throws IOException Any problem during remote communication.
   */
  public ServerResponse sendRequest(ClientRequest request, IResponseStreamHandler responseStreamHandler) throws IOException
  {
    HttpURLConnection connection;
    ServerResponse response;

    connection = this.openHttpConnection(request.getURL());
    try
    {
      connection.setRequestMethod(request.getHttpMethod().getName());
      connection.setReadTimeout(this.getReadTimeout());
      connection.setConnectTimeout(this.getConnectTimeout());
      connection.setDoInput(true);
      this.applyUserAgent(connection);
      this.applyRequestHeaders(connection, this.getHeaders());
      this.applyRequestHeaders(connection, request.getHeaders());
      if (request.hasContentData())
      {
        connection.setDoOutput(true);
        connection.setRequestProperty(RequestHeaderNames.CONTENT_LENGTH, Integer.toString(request.getContentLength()));
        this.sendRequestBody(connection, request);
      }
      response = this.retrieveResponse(connection, responseStreamHandler);
    }
    finally
    {
      connection.disconnect();
    }
    return response;
  } // sendRequest() 

  // -------------------------------------------------------------------------

  public int getReadTimeout()
  {
    return this.readTimeout;
  } // getReadTimeout() 

  // -------------------------------------------------------------------------

  public void setReadTimeout(int readTimeout)
  {
    this.readTimeout = readTimeout;
  } // setReadTimeout() 

  // -------------------------------------------------------------------------

  public int getConnectTimeout()
  {
    return this.connectTimeout;
  } // getConnectTimeout() 

  // -------------------------------------------------------------------------

  public void setConnectTimeout(int connectTimeout)
  {
    this.connectTimeout = connectTimeout;
  } // setConnectTimeout() 

  // -------------------------------------------------------------------------

  public String getUserAgentName()
  {
    return userAgentName;
  } // getUserAgentName() 

  // -------------------------------------------------------------------------

  public void setUserAgentName(String userAgentName)
  {
    this.userAgentName = userAgentName;
  } // setUserAgentName() 

  public boolean hasProxy()
  {
    return this.getProxy() != null;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected HttpURLConnection openHttpConnection(URL url) throws IOException
  {
    if (this.hasProxy())
    {
      return (HttpURLConnection)url.openConnection(this.getProxy());
    }
    return (HttpURLConnection)url.openConnection();
  }

  protected void sendRequestBody(HttpURLConnection connection, ClientRequest request) throws IOException
  {
    DataOutputStream outStream;

    outStream = new DataOutputStream(connection.getOutputStream());
    try
    {
      outStream.write(request.getContentData());
      outStream.flush();
    }
    finally
    {
      outStream.close();
    }
  } // sendRequestBody() 

  // -------------------------------------------------------------------------

  protected void applyUserAgent(HttpURLConnection connection)
  {
    String agentName;
    agentName = connection.getRequestProperty(RequestHeaderNames.USER_AGENT) + " " + this.getUserAgentName();
    connection.addRequestProperty(RequestHeaderNames.USER_AGENT, agentName);
  } // applyUserAgent()

  // -------------------------------------------------------------------------

  protected void applyRequestHeaders(HttpURLConnection connection, Headers headerDef)
  {
    String[] headerValues;

    for (String headerName : headerDef.getHeaderNames())
    {
      headerValues = headerDef.getHeaderValues(headerName);
      if (SU.isNullOrEmpty(headerValues))
      {
        this.getLogger().logDebug("Skipped empty request header: {0}", headerName);
      }
      else
      {
        // TODO Check what separator is to be used!!!
        connection.setRequestProperty(headerName, SU.asString(headerValues, " "));
      }
    }
  } // applyRequestHeaders() 

  // -------------------------------------------------------------------------

  protected ServerResponse retrieveResponse(final HttpURLConnection connection, final IResponseStreamHandler responseStreamHandler) throws IOException
  {
    InputStream inStream;
    final ServerResponse response;
    int responseCode;
    IResponseStreamHandler inStreamHandler;

    //    responseCode = HttpUtil.current().getResponseCode(connection);
    responseCode = connection.getResponseCode();
    response = new ServerResponse(responseCode, connection.getResponseMessage());
    response.addAllHeaders(connection.getHeaderFields());

    if (this.isSupposedToHaveContent(response))
    {
      inStream = (responseCode < 400) ? connection.getInputStream() : connection.getErrorStream();
      if (inStream != null)
      {
        inStreamHandler = responseStreamHandler;
        if (inStreamHandler == null)
        {
          inStreamHandler = new IResponseStreamHandler()
          {
            @Override
            public void handleStream(InputStream stream) throws IOException
            {
              response.appendContent(stream);
            }
          };
        }
        inStreamHandler.handleStream(inStream);
      }
    }
    return response;
  } // retrieveResponse() 

  // -------------------------------------------------------------------------

  protected boolean isSupposedToHaveContent(ServerResponse response)
  {
    int contentLength;
    boolean isChunked;

    contentLength = SU.asInteger(response.getHeader(ResponseHeaderNames.CONTENT_LENGTH), 0);
    isChunked = "chunked".equalsIgnoreCase(response.getHeader(ResponseHeaderNames.TRANSFER_ENCODING));

    return (this.canHaveContent(response.getResponseCode()) && ((contentLength > 0) || isChunked));
  }

  // -------------------------------------------------------------------------

  protected boolean canHaveContent(int responseCode)
  {
    switch (responseCode)
    {
      case 204 :
      case 205 :
      case 301 :
      case 302 :
      case 303 :
      case 304 :
      case 307 :
      case 401 :
        return false;
      default :
        return true;
    }
  } // canHaveContent() 

  // -------------------------------------------------------------------------

  protected Headers getHeaders()
  {
    return headers;
  } // getHeaders() 

  // -------------------------------------------------------------------------

  protected Logger getLogger()
  {
    return LoggerProvider.getComponentLogger();
  } // getLogger() 

  // -------------------------------------------------------------------------

} // class UserAgent 
