// ===========================================================================
// CONTENT  : CLASS SecurityUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 18/08/2014
// HISTORY  :
//  18/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.security;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;

import org.pfsw.bif.text.IStringPair;
import org.pfsw.security.crypto.CryptoConstants;
import org.pfsw.text.StringUtil;
import org.pfsw.util.Base64Converter;
import org.pfsw.util.SysUtil;

/**
 * This singleton supports utility methods around various security aspects.
 * It has a protected default constructor, so it can be extended!
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class SecurityUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();
  private static final SysUtil SYSU = SysUtil.current();

  public static final String BASIC_CREDENTIALS_SEPARATOR = ":";

  public static final String AUTHENTICATION_SCHEME_BASIC = "Basic";

  public static final String AUTHENTICATION_SCHEME_DIGEST = "Digest";

  public static final String AUTHENTICATION_SCHEME_BEARER = "Bearer";

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static SecurityUtil soleInstance = new SecurityUtil();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================

  // =========================================================================
  // CLASS METHODS
  // =========================================================================
  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static SecurityUtil current()
  {
    return soleInstance;
  } // current() 

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  protected SecurityUtil()
  {
    super();
  } // SecurityUtil() 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Encodes the given username and password according to the BASIC authentication
   * scheme specification (RFC-2617).
   * <br>
   * That is base64(username:password). 
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @return The base64 encoded concatenated username password.
   */
  public String encodeBASICCredentials(String username, String password)
  {
    String both;

    both = username + BASIC_CREDENTIALS_SEPARATOR + password;
    return Base64Converter.encode(both);
  } // encodeBASICCredentials() 

  // -------------------------------------------------------------------------

  /**
   * Encodes the given username and password according to the BASIC authentication
   * scheme specification (RFC-2617).
   * <br>
   * That is base64(username:password). 
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return The base64 encoded concatenated username and password.
   */
  public String encodeBASICCredentials(String username, String password, Charset charset)
  {
    String both;

    both = username + BASIC_CREDENTIALS_SEPARATOR + password;
    return Base64Converter.encodeToString(both.getBytes(charset));
  } // encodeBASICCredentials() 

  // -------------------------------------------------------------------------

  /**
   * Decodes and splits up the value presented by a BASIC authentication 
   * header as to RFC-2617.
   * 
   * @param basicCredentials The base64 encoded concatenated username password string (must not be null). 
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return A string pair where the first string contains the username and the second the password.
   * @throws IllegalArgumentException If the given string cannot be decoded or cannot be split into two sub-values.
   */
  public IStringPair decodeBASICCredentials(String basicCredentials, Charset charset)
  {
    byte[] bytes;
    String decoded;
    IStringPair pair;

    bytes = Base64Converter.decode(basicCredentials);
    decoded = new String(bytes, charset);
    pair = SU.splitStringPair(decoded, BASIC_CREDENTIALS_SEPARATOR);
    if (SU.isNullOrBlank(pair.getString1()) || SU.isNullOrBlank(pair.getString2()))
    {
      throw new IllegalArgumentException("Failed to split up the given credentials.");
    }
    return pair;
  } // decodeBASICCredentials() 

  // -------------------------------------------------------------------------

  /**
   * Decodes and splits up the value presented by a BASIC authentication 
   * header as to RFC-2617.
   * 
   * @param basicCredentials The base64 encoded concatenated username password string (must not be null). 
   * @return A string pair where the first string contains the username and the second the password.
   * @throws IllegalArgumentException If the given string cannot be decoded or cannot be split into two sub-values.
   */
  public IStringPair decodeBASICCredentials(String basicCredentials)
  {
    String decoded;
    IStringPair pair;

    decoded = Base64Converter.decodeToString(basicCredentials);
    pair = SU.splitStringPair(decoded, BASIC_CREDENTIALS_SEPARATOR);
    if (SU.isNullOrBlank(pair.getString1()) || SU.isNullOrBlank(pair.getString2()))
    {
      throw new IllegalArgumentException("Failed to split up the given credentials.");
    }
    return pair;
  } // decodeBASICCredentials() 

  // -------------------------------------------------------------------------

  /**
   * Creates the full authorization header value for the HTTP BASIC authentication
   * scheme as defined in RFC-2617.
   * Username and password will be concatenated and base64 encoded.
   * 
   * @param username The username (must not be null).
   * @param password The password (must not be null)
   * @return The complete header value (e.g. "Basic Q29yb2xpbmE6N1J1YmJlcmJvYXQyNQ==")
   */
  public String createBASICAuthorization(String username, String password)
  {
    String credentials;

    credentials = this.encodeBASICCredentials(username, password);
    return AUTHENTICATION_SCHEME_BASIC + " " + credentials;
  } // createBASICAuthorization() 

  // -------------------------------------------------------------------------

  /**
   * Creates the full authorization header value for the BEARER authentication
   * scheme as defined in RFC-6750.
   * 
   * @param token The token that will be base64 encoded by this method (must not be null).
   * @return The complete header value (e.g. "Bearer cy9aJsKnPX1+SDIjJz9ERlYoLyY9U0HDvFNEQEo34oKsesOfZGpkag==")
   */
  public String createBEARERAuthorization(byte[] token)
  {
    String encoded;

    encoded = Base64Converter.encodeToString(token);
    return AUTHENTICATION_SCHEME_BEARER + " " + encoded;
  } // createBEARERAuthorization() 

  // -------------------------------------------------------------------------

  /**
   * Creates the full authorization header value for the BEARER authentication
   * scheme as defined in RFC-6750.
   * 
   * @param token The token that will be base64 encoded by this method (must not be null).
   * @param charset The charset to be used for the string encoding (must not be null).
   * @return The complete header value (e.g. "Bearer cy9aJsKnPX1+SDIjJz9ERlYoLyY9U0HDvFNEQEo34oKsesOfZGpkag==")
   */
  public String createBEARERAuthorization(String token, Charset charset)
  {
    return this.createBEARERAuthorization(token.getBytes(charset));
  } // createBEARERAuthorization() 

  // -------------------------------------------------------------------------

  /**
   * Read all certificates from the given PKCS#7 stream.
   * The given stream will be closed in any case after this call.
   * 
   * @param inStream A stream that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The certificates read from the given PKCS#7 stream.
   * @throws CertificateException For any case of format or IO errors.
   */
  public Collection<X509Certificate> readX509CertificatesFromPKCS7(final InputStream inStream) throws CertificateException
  {
    CertificateFactory certFactory;
    Collection<? extends Certificate> certificates;
    Collection<X509Certificate> result;

    try
    {      
      certFactory = CertificateFactory.getInstance(CryptoConstants.CERT_STANDARD_X509);
      certificates = certFactory.generateCertificates(inStream);
      result = new ArrayList<X509Certificate>(certificates.size());
      for (Certificate certificate : certificates)
      {
        result.add((X509Certificate)certificate);
      }
      return result;
    }
    finally
    {
      SYSU.close(inStream);
    }
  } // readX509CertificatesFromPKCS7() 

  // -------------------------------------------------------------------------

  /**
   * Reads a certificate from the given PKCS#7 stream.
   * 
   * @param inStream A stream that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The (first) X.509 certificate read from the given PKCS#7 stream.
   * @throws CertificateException For any case of format or IO errors.
   */
  public X509Certificate readX509CertificateFromPKCS7(final InputStream inStream) throws CertificateException
  {
    Collection<X509Certificate> result;

    result = this.readX509CertificatesFromPKCS7(inStream);
    if (result.isEmpty())
    {
      return null;
    }
    return result.iterator().next();
  } // readX509CertificatesFromPKCS7() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Reads a certificate from the given PKCS#7 file.
   * 
   * @param file A file that is expected to contain X.509 certificates in PKCS#7 format.
   * @return The (first) X.509 certificate read from the given PKCS#7 file.
   * @throws CertificateException For any case of format or IO errors.
   */
  public X509Certificate readX509CertificateFromPKCS7(final File file) throws CertificateException
  {
    FileInputStream inStream;
    
    try
    {
      inStream = new FileInputStream(file);
    }
    catch (FileNotFoundException ex)
    {
      throw new CertificateException("Certificate file " + file.getAbsolutePath() + " not found", ex);
    }
    return this.readX509CertificateFromPKCS7(inStream);
  } // readX509CertificatesFromPKCS7() 
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================

} // class SecurityUtil 
