// ===========================================================================
// CONTENT  : CLASS JsonObjectAccessor
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.1 - 16/02/2017
// HISTORY  :
//  08/01/2017  mdu  CREATED
//  16/02/2017  mdu   added --> getRequiredStringArrayValue(), getOptionalStringArrayValue()
//
// Copyright (c) 2017, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

import java.math.BigDecimal;

/**
 * This is a wrapper around a single JsonObject. It supports access to the 
 * objects's field values in a convenient way and supports optional and required
 * field access as well as implicit type conversion.
 * All access problems are signaled by {@link JsonAccessException} which is a RuntimeException.
 *
 * @author Manfred Duchrow
 * @version 1.1
 */
public class JsonObjectAccessor
{
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final JsonObject jsonObject;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public JsonObjectAccessor(final JsonObject jsonObject)
  {
    super();
    this.jsonObject = jsonObject;
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public JsonObject getJsonObject()
  {
    return this.jsonObject;
  }
  
  public JsonObjectAccessor getRequiredJsonObjectAccessorValue(final String name)
  {
    JsonObject object;
    
    object = this.getRequiredJsonObjectValue(name);
    return new JsonObjectAccessor(object);
  } 
  
  public JsonObject getRequiredJsonObjectValue(final String name)
  {
    return this.getRequiredTypedValue(name, JsonObject.class);
  } 
  
  public JsonArray getRequiredJsonArrayValue(final String name)
  {
    return this.getRequiredTypedValue(name, JsonArray.class);
  } 
  
  public String getRequiredStringValue(final String name)
  {
    return this.getRequiredTypedValue(name, String.class);
  } 
  
  public String[] getRequiredStringArrayValue(final String name)
  {
    return this.getStringArray(name, true);
  }
  
  public Boolean getRequiredBooleanValue(final String name)
  {
    return this.getRequiredTypedValue(name, Boolean.class);
  } 
  
  public Integer getRequiredIntegerValue(final String name)
  {
    return this.getRequiredTypedValue(name, Integer.class);
  } 

  public Long getRequiredLongValue(final String name)
  {
    return this.getRequiredTypedValue(name, Long.class);
  }

  public BigDecimal getRequiredBigDecimalValue(final String name)
  {
    return this.getRequiredTypedValue(name, BigDecimal.class);
  }
  
  public Object getRequiredRawValue(final String name)
  {
    return this.getRawValueOf(name, true);
  }  

  public JsonObjectAccessor getOptionalJsonObjectAccessorValue(final String name)
  {
    JsonObject object;
    
    object = this.getOptionalJsonObjectValue(name);
    if (object == null)
    {
      return null;
    }
    return new JsonObjectAccessor(object);
  } 
  
  public JsonObject getOptionalJsonObjectValue(final String name)
  {
    return this.getOptionalTypedValue(name, JsonObject.class);
  } 
  
  public JsonArray getOptionalJsonArrayValue(final String name)
  {
    return this.getOptionalTypedValue(name, JsonArray.class);
  } 
  
  public String getOptionalStringValue(final String name)
  {
    return this.getOptionalTypedValue(name, String.class);
  } 
  
  public String[] getOptionalStringArrayValue(final String name)
  {
    return this.getStringArray(name, false);
  } 
  
  public Boolean getOptionalBooleanValue(final String name)
  {
    return this.getOptionalTypedValue(name, Boolean.class);
  } 
  
  public Integer getOptionalIntegerValue(final String name)
  {
    return this.getOptionalTypedValue(name, Integer.class);
  } 

  public Long getOptionalLongValue(final String name)
  {
    return this.getOptionalTypedValue(name, Long.class);
  }

  public BigDecimal getOptionalBigDecimalValue(final String name)
  {
    return this.getOptionalTypedValue(name, BigDecimal.class);
  }
  
  public Object getOptionalRawValue(final String name)
  {
    return this.getRawValueOf(name, false);
  }
  
  public <T> T getOptionalTypedValue(final String name, Class<T> type)
  {
    return this.getTypedValueOf(name, type, false);
  }

  public <T> T getRequiredTypedValue(final String name, Class<T> type)
  {
    return this.getTypedValueOf(name, type, true);
  }
  
  public <T> T getTypedValueOf(final String name, final Class<T> type, final boolean isRequired)
  {
    T value;

    try
    {
      if (type == Long.class)
      {
        value = this.getLongValueOf(name, type);
      }
      else
      {
        value = this.getJsonObject().getValueOfType(name, type);
      }
    }
    catch (Exception ex)
    {
      throw new JsonAccessException(ex, "Type conversion problem for value of \"%s\". The value is not %s", name, type.getSimpleName());
    }
    if ((value == null) && isRequired)
    {
      this.signalRequiredFieldMissing(name);
    }
    return value;
  }

  public Object getRawValueOf(final String name, final boolean isRequired)
  {
    Object value;
    
    value = this.getJsonObject().get(name);
    if ((value == null) && isRequired)
    {
      this.signalRequiredFieldMissing(name);
    }
    return value;
  }
  
  public void signalRequiredFieldMissing(String fieldName) throws JsonAccessException
  {
    throw new JsonAccessException("Required field missing: \"%s\"", fieldName);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected String[] getStringArray(final String name, boolean isRequired) 
  {
    JsonArray jsonArray;
    String[] strings;
    
    jsonArray = this.getTypedValueOf(name, JsonArray.class, isRequired);
    if (jsonArray == null)
    {
      return null;
    }
    strings = new String[jsonArray.size()];
    for (int i = 0; i < strings.length; i++)
    {
      strings[i] = jsonArray.getString(i);
    }
    return strings;
  }
  
  protected <T> T getLongValueOf(final String name, final Class<T> type)
  {
    Object value;
    Integer intValue;

    value = this.getJsonObject().get(name);
    if (value == null)
    {
      return null;
    }
    if (value instanceof Long)
    {
      return (T)value;
    }
    if (value instanceof Integer)
    {
      intValue = (Integer)value;
      return (T)Long.valueOf(intValue.longValue());
    }
    throw new ClassCastException("Field named <" + name + "> is no " + type.getName() + ". It is a " + value.getClass().getName());
  }
}
