// ===========================================================================
// CONTENT  : CLASS OrderedProperties
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.1 - 16/11/2014
// HISTORY  :
//  17/09/2004  mdu  CREATED
//  16/11/2014  mdu  refactored --> generic types
//
// Copyright (c) 2004-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pfsw.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;

import org.pfsw.bif.text.IMutableStringPair;

/**
 * Represents a properties collection with empty lines and comments.
 * The order will be preserved.
 * <b>DO NOT USE THE NORMAL METHODS INHERITED FROM java.util.Hashtable
 * except those that are explicitly overridden in this class!</b>
 *
 * @author Manfred Duchrow
 * @version 2.1
 */
public class OrderedProperties<T extends IMutableStringPair> extends Properties
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 4168763680301505984L;

  protected static final int INITIAL_CAPACITY		= 50 ;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private List<T> elements = null ;
  protected List<T> getElements() { return elements ; }
  protected void setElements( List<T> newValue ) { elements = newValue ; }
  
  private Map<String, Integer> propertyIndex = null ;
  protected Map<String, Integer> getPropertyIndex() { return propertyIndex ; }
  protected void setPropertyIndex( Map<String, Integer> newValue ) { propertyIndex = newValue ; } 
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public OrderedProperties()
  {
    this(INITIAL_CAPACITY);
  } // OrderedProperties() 

  // -------------------------------------------------------------------------

  /**
   * Initialize the new instance with an initial capacity.
   */
  public OrderedProperties(int initialCapacity)
  {
    super();
    this.setElements(this.newElementsList(initialCapacity));
    this.setPropertyIndex(new HashMap<String, Integer>(initialCapacity));
  } // OrderedProperties() 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a list containing all properties names in the order they 
   * have been added or read from a stream.
   */
  public List<String> nameList()
  {
    List<String> nameList;
    
    nameList = new ArrayList<String>(this.size());
    for (T property : this.getElements())
    {      
      if (this.isValidProperty(property))
      {
        nameList.add(property.getString1());
      }
    }
    return nameList;
  } // nameList() 
  
  // -------------------------------------------------------------------------

  /**
   * Returns an array containing the properties names in the order they 
   * have been added or read from a stream.
   */
  public String[] getPropertyNames()
  {
    List<String> names;

    names = this.nameList();
    return names.toArray(new String[names.size()]);
  } // getPropertyNames() 

  // -------------------------------------------------------------------------

  /**
   * Returns how many properties are contained. 
   * That does NOT include empty and comment lines!
   * @see #numberOfElements()
   */
  @Override
  public synchronized int size()
  {
    int count = 0;
    
    for (T property : this.getElements())
    {      
      if (this.isValidProperty(property))
      {
        count++;
      }
    }
    return count;
  } // size() 

  // -------------------------------------------------------------------------
  
  /**
   * Returns how many elements are contained. That includes empty and comment lines!
   * @see #size()
   */
  public int numberOfElements() 
  {
    return this.getElements().size();
  } // numberOfElements()
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns whether or not this contains any properties.
   * Empty and comment lines are ignored here.
   * @see #hasContent()
   */
  @Override
  public synchronized boolean isEmpty()
  {
    return this.size() < 1;
  } // isEmpty() 
  
  // -------------------------------------------------------------------------

  /**
   * Returns whether or not any data is contained.
   * That includes empty and comment lines are ignored here.
   * So this method can return true even if no property is defined.
   * @see #isEmpty()
   */
  public boolean hasContent() 
  {
    return this.numberOfElements() > 0;
  } // hasContent()
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the value of the property with the given name or null if it is
   * not found.
   * 
   * @param name The name of the property to look for
   */
  @Override
  public String getProperty(String name)
  {
    T property;

    property = this.findProperty(name);
    if (this.isNullOrDeleted(property))
    {
      return null;
    }
    return property.getString2();
  } // getProperty() 

  // -------------------------------------------------------------------------

  /**
   * Modifies the value or adds the property specified by the given name.
   * Returns true the previous value if the property was found and modified 
   * or null if the property was added.
   * 
   * @param name The name of the property to set
   * @param value The new value to be set
   */
  @Override
  public synchronized Object setProperty(String name, String value)
  {
    T property;
    String oldValue;

    property = this.findProperty(name);
    if (property == null)
    {
      this.appendProperty(name, value);
      return null;
    }
    oldValue = property.getString2();
    property.setString2(value);
    return oldValue;
  } // setProperty() 

  // -------------------------------------------------------------------------

  /**
   * Deletes the property with the specified name.
   * Returns true if the property was deleted, otherwise false.
   * 
   * @param name The name of the property to delete
   */
  public boolean deleteProperty(String name)
  {
    return this.remove(name) != null;
  } // deleteProperty() 

  // -------------------------------------------------------------------------

  /**
   * Removes the key (and its corresponding value) from this hashtable. 
   * This method does nothing if the key is not found.
   * 
   * @return the value to which the key had been mapped, or null if the key did not have a mapping.
   */
  @Override
  public synchronized Object remove(Object key)
  {
    T property;

    if (!(key instanceof String))
    {
      return null;
    }

    property = this.findProperty((String)key);
    if (property == null)
    {
      return null;
    }
    this.getPropertyIndex().remove(property.getString1());
    this.getElements().remove(property);
    return property.getString2();
  } // remove() 

  // -------------------------------------------------------------------------

  /**
   * Searches for the property with the specified key in this property list. 
   * If the key is not found in this property list the method returns the 
   * default value argument.
   */
  @Override
  public String getProperty(String key, String defaultValue)
  {
    T property;

    property = this.findProperty(key);
    if (property == null)
    {
      return defaultValue;
    }
    return property.getString2();
  } // getProperty() 

  // -------------------------------------------------------------------------

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public void list(PrintStream out)
  {
    throw new UnsupportedOperationException(this.getClass().getName() + ".list()");
  } // list() 

  // -------------------------------------------------------------------------

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public void list(PrintWriter out)
  {
    throw new UnsupportedOperationException(this.getClass().getName() + ".list()");
  } // list() 

  // -------------------------------------------------------------------------

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public synchronized void load(InputStream inStream) throws IOException
  {
    throw new UnsupportedOperationException(this.getClass().getName() + ".load()");
  } // load() 

  // -------------------------------------------------------------------------

  /**
   * Returns an enumeration of the properties names in the order they have been
   * added or read from a stream.
   * <p>
   * <b>Be aware that the returned enumeration is NOT backed by this 
   * OrderedProperties object!</b>  
   */
  @Override
  public Enumeration<String> propertyNames()
  {
    Vector<String> names;

    names = new Vector<String>(this.nameList());
    return names.elements();
  } // propertyNames() 

  // -------------------------------------------------------------------------
  
  @Override
  public Set<String> stringPropertyNames()
  {
    Set<String> nameSet;
    
    nameSet = new OrderedSet<String>(this.size());
    nameSet.addAll(this.nameList());
    return nameSet;
  } // stringPropertyNames() 
  
  // -------------------------------------------------------------------------

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @SuppressWarnings("deprecation")
  @Override
  public synchronized void save(OutputStream out, String header)
  {
    throw new UnsupportedOperationException(this.getClass().getName() + ".save()");
  } // save() 

  // -------------------------------------------------------------------------

  /**
   * NOT SUPPORTED METHOD!
   * @throws UnsupportedOperationException
   */
  @Override
  public synchronized void store(OutputStream out, String header) throws IOException
  {
    throw new UnsupportedOperationException(this.getClass().getName() + ".store()");
  } // store() 

  // -------------------------------------------------------------------------

  /**
   * Removes all properties 
   */
  @Override
  public synchronized void clear()
  {
    this.getElements().clear();
    this.getPropertyIndex().clear();
  } // clear() 

  // -------------------------------------------------------------------------

  /**
   * Returns all keys of this properties list in the order they have been added.
   */
  @SuppressWarnings({ "rawtypes", "unchecked" })
  @Override
  public synchronized Enumeration keys()
  {
    return this.propertyNames();
  } // keys() 

  // -------------------------------------------------------------------------

  /**
   * Returns all keys of this properties list in the order they have been added.
   */
  @SuppressWarnings({ "unchecked", "rawtypes" })
  @Override
  public Set keySet()
  {
    return this.stringPropertyNames();
  } // keySet() 

  // -------------------------------------------------------------------------

  /**
   * Returns all entries in the order they have been added
   */
  @SuppressWarnings({ "rawtypes", "unchecked" })
  @Override
  public Set entrySet()
  {
    OrderedMap entries;
    Iterator<T> iter;
    T element;

    entries = new OrderedMap();
    iter = this.getElements().iterator();
    while (iter.hasNext())
    {
      element = iter.next();
      if (this.isValidProperty(element))
      {
        entries.put(element.getString1(), element.getString2());
      }
    }
    return entries.entrySet();
  } // entrySet() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Appends the property specified by the given name with the given value..
   * Returns true if the property was successfully added at the end, 
   * otherwise false.
   * 
   * @param name The name of the property to append
   * @param value The new value to be append
   */
  protected boolean appendProperty(String name, String value)
  {
    T property;

    property = this.newProperty(name, value);
    this.addElement(property);
    return true;
  } // appendProperty() 

  // -------------------------------------------------------------------------

  /**
   * Create a new property object with the given name and value. Subclasses
   * may override this to implement their own property class.
   */
  protected T newProperty(String name, String value)
  {
    return (T)new NamedText(name, value);
  } // newProperty() 

  // -------------------------------------------------------------------------

  protected void addElement(T element)
  {
    Integer index;

    index = new Integer(this.getElements().size());
    this.getElements().add(element);
    this.registerProperty(element, index);
  } // addElement() 

  // -------------------------------------------------------------------------

  protected void registerProperty(T property, Integer index)
  {
    this.getPropertyIndex().put(property.getString1(), index);
  } // registerProperty() 

  // -------------------------------------------------------------------------

  protected T findProperty(String key)
  {
    Integer index;

    index = this.indexOfProperty(key);
    if (index == null)
    {
      return null;
    }
    return this.propertyAt(index);
  } // findProperty() 

  // -------------------------------------------------------------------------

  protected Integer indexOfProperty(String key)
  {
    return this.getPropertyIndex().get(key);
  } // indexOfProperty() 

  // -------------------------------------------------------------------------

  protected T propertyAt(Integer index)
  {
    return this.propertyAt(index.intValue());
  } // propertyAt() 

  // ------------------------------------------------------------------------- 

  protected T propertyAt(int index)
  {
    return this.getElements().get(index);
  } // propertyAt() 

  // -------------------------------------------------------------------------

  /**
   * Returns true if the given property is null or marked as deleted.
   * Subclasses may override this method to enhance behavior.
   * 
   * @param property The property to check
   * @return true if the given property is null or marked as deleted
   */
  protected boolean isNullOrDeleted(T property)
  {
    return (property == null);
  } // isNullOrDeleted() 

  // -------------------------------------------------------------------------

  protected boolean isValidProperty(T property)
  {
    return true;
  } // isValidProperty() 

  // -------------------------------------------------------------------------
  
  protected List<T> newElementsList(int initialCapacity)
  {
    return new ArrayList<T>(initialCapacity);
  }

} // class OrderedProperties 
