// ===========================================================================
// CONTENT  : CLASS ReleaseState
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 25/05/2015
// HISTORY  :
//  25/05/2015  mdu  CREATED
//
// Copyright (c) 2015, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Serializable;
import java.util.concurrent.atomic.AtomicBoolean;

import org.pfsw.bif.lifecycle.IReleasable;

/**
 * A simple handler of the released state of an object that might implement the
 * {@link IReleasable} interface.
 * <br>
 * It can be used by sublclassing or as a delegate property, whatever suits better.
 * <br>
 * This implementation is thread-safe. 
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class ReleaseState implements IReleasable, Serializable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = -3707502941883821353L;
  
  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final AtomicBoolean isReleased = new AtomicBoolean(false);

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Creates a new instance this has the state "not released".
   */
  public ReleaseState()
  {
    super();
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns true if this object has been released.
   */
  public boolean isReleased()
  {
    return this.isReleased.get();
  }

  /**
   * Checks if this object has been released already and throws an IllegalStateException
   * if that is the case. Otherwise it silently returns without any action.
   * 
   * @throws IllegalStateException If the object has been released.
   */
  public void checkState()
  {
    if (this.isReleased())
    {
      throw new IllegalStateException("The resource of this authorization handler have been already released! " + this.getClass().getName());
    }
  }

  /**
   * Changes the internal state to "released".
   */
  @Override
  public void release()
  {
    this.setIsReleased(true);
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected void setIsReleased(boolean bool)
  {
    this.isReleased.set(bool);
  }
}
