// ===========================================================================
// CONTENT  : CLASS StreamCopy
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 12/02/2017
// HISTORY  :
//  12/02/2017  mdu  CREATED
//
// Copyright (c) 2017, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.pfsw.bif.callback.IProgressListener;

/**
 * This class supports convenient stream copying.
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class StreamCopy
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static boolean DEFAULT_CLOSE_AFTER_COPY = true;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private int bufferSize = IOUtil.DEFAULT_BUFFER_SIZE;
  private boolean closeStreamsAfterCopy = true;
  private IProgressListener progressListener = null;
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public StreamCopy()
  {
    this(IOUtil.DEFAULT_BUFFER_SIZE, DEFAULT_CLOSE_AFTER_COPY);
  }

  public StreamCopy(IProgressListener listener)
  {
    this();
    this.setProgressListener(listener);
  }

  public StreamCopy(int bufferSize)
  {
    this(bufferSize, DEFAULT_CLOSE_AFTER_COPY);
  }
  
  public StreamCopy(boolean closeStreamsAfterCopy)
  {
    this(IOUtil.DEFAULT_BUFFER_SIZE, closeStreamsAfterCopy);
  }
  
  public StreamCopy(int bufferSize, IProgressListener listener)
  {
    this(bufferSize, DEFAULT_CLOSE_AFTER_COPY, listener);
  }
  
  public StreamCopy(boolean closeStreamsAfterCopy, IProgressListener listener)
  {
    this(IOUtil.DEFAULT_BUFFER_SIZE, closeStreamsAfterCopy, listener);
  }
  
  public StreamCopy(int bufferSize, boolean closeStreamsAfterCopy)
  {
    this(bufferSize, closeStreamsAfterCopy, null);
  }
  
  public StreamCopy(int bufferSize, boolean closeStreamsAfterCopy, IProgressListener listener)
  {
    super();
    this.setBufferSize(bufferSize);
    this.setCloseStreamsAfterCopy(closeStreamsAfterCopy);
    this.setProgressListener(listener);
  }

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public int getBufferSize()
  {
    return this.bufferSize;
  }
  
  public void setBufferSize(int bufferSize)
  {
    this.bufferSize = bufferSize;
  }
  
  public boolean isCloseStreamsAfterCopy()
  {
    return this.closeStreamsAfterCopy;
  }
  
  public void setCloseStreamsAfterCopy(boolean closeStreamsAfterCopy)
  {
    this.closeStreamsAfterCopy = closeStreamsAfterCopy;
  }
  
  public IProgressListener getProgressListener()
  {
    return this.progressListener;
  }
  
  public void setProgressListener(IProgressListener listener)
  {
    this.progressListener = listener;
  }
  
  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the underlying defined size in bytes.
   * <br>
   * Depending on the value of 'closeStreamsAfterCopy' the two streams will
   * be closed at the end of this method execution.
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @return The number of bytes copied.
   */
  public long copy(InputStream inStream, OutputStream outStream) throws IOException
  {
    return this.copy(inStream, outStream, null);
  }
  
  /**
   * Copies all data from the input stream to the output stream using
   * a buffer of the underlying defined size in bytes.
   * <br>
   * Depending on the value of 'closeStreamsAfterCopy' the two streams will
   * be closed at the end of this method execution.
   * 
   * @param inStream The source stream.
   * @param outStream The target stream.
   * @param limit The maximum number of bytes that will be copied. If exceeded an IOException will be thrown.
   *   Setting this parameter to null, zero or a negative value will specify no limit.
   * @return The number of bytes copied.
   */
  public long copy(InputStream inStream, OutputStream outStream, Long limit) throws IOException
  {
    byte[] buffer;
    int count;
    int readLen;
    long copiedBytes = 0L;
    Long copyLimit;
    boolean goOn = true;
    int bufSize;
    
    bufSize = this.getBufferSize();
    buffer = new byte[bufSize];
    copyLimit = (limit == null) ? -1 : limit.longValue();
    try
    {
      if (copyLimit <= 0) // No limit?
      {
        readLen = bufSize;
      }
      else
      {
        // If the limit is less than the buffer size, ensure only
        // the limited number of bytes will be read.
        readLen = (copyLimit < bufSize) ? copyLimit.intValue() : bufSize;
      }
      if (this.getProgressListener() != null)
      {
        goOn = this.getProgressListener().processed(IProgressListener.UNDEFINED_SIZE, copiedBytes);
      }
      if (goOn)
      {        
        count = inStream.read(buffer, 0, readLen);
        while (goOn && (count > -1))
        {
          outStream.write(buffer, 0, count);
          copiedBytes += count;
          if (copyLimit > 0) // Any limit at all?
          {
            if (copiedBytes >= copyLimit)
            {
              // Limit reached - get out of here
              throw new IOException("Copy limit of " + copyLimit + " bytes reached.");
            }
            Long remainingLimit = copyLimit - copiedBytes;
            readLen = (remainingLimit < bufSize) ? remainingLimit.intValue() : bufSize;
          }
          if (this.getProgressListener() != null)
          {
            goOn = this.getProgressListener().processed(IProgressListener.UNDEFINED_SIZE, copiedBytes);
          }
          if (goOn)
          {            
            count = inStream.read(buffer, 0, readLen);
          }
        }
      }
    }
    finally
    {
      if (this.isCloseStreamsAfterCopy())
      {        
        this.ioUtil().close(inStream);
        this.ioUtil().close(outStream);
      }
    }
    return copiedBytes;
  }

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected IOUtil ioUtil()
  {
    return IOUtil.current();
  }
}
