// ===========================================================================
// CONTENT  : CLASS TempFilesManager
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 30/11/2014
// HISTORY  :
//  30/11/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.file;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicInteger;

import org.pfsw.bif.logic.BooleanOperation;
import org.pfsw.text.StringFilterCollection;
import org.pfsw.text.StringPattern;
import org.pfsw.text.StringUtil;

/**
 * An instance of this class is maintaining an arbitrary number of temporary file
 * in a dedicated folder of the file system (on which it must have write access).
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class TempFilesManager
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final FileUtil FU = FileUtil.current();
  private static final StringUtil SU = StringUtil.current();

  public static final String TIMESTAMP_FORMAT = "yyyy-MM-dd'T'HH-mm-ss";
  public static final String DEFAULT_SUFFIX = ".tmp";
  public static final String NAME_PART_SEPARATOR = "__";
  public static final int DEFAULT_BUFFER_SIZE = 8096;

  private static final String TIMESTAMP_SEARCH_PATTERN = "*" + NAME_PART_SEPARATOR + "????-??-??T??-??-??" + NAME_PART_SEPARATOR + "*";

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final File tempFolder;
  private final AtomicInteger nextNumericId = new AtomicInteger(0);
  private String fileNamePrefix = null;
  private String fileNameSuffix = DEFAULT_SUFFIX;
  private Long fileSizeLimit = null; // Null means no limit
  private final String uniqueId;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public TempFilesManager(File tempFolder) throws IOException
  {
    this(tempFolder, false);
  } // TempFilesManager() 

  // -------------------------------------------------------------------------

  public TempFilesManager(File tempFolder, boolean requiresInstanceUniqueness) throws IOException
  {
    super();
    this.tempFolder = tempFolder;
    this.checkTempFolder();
    if (requiresInstanceUniqueness)
    {
      this.uniqueId = this.calculateInstanceUniqueId();
    }
    else
    {
      this.uniqueId = null;
    }
  } // TempFilesManager() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  public String getFileNamePrefix()
  {
    return fileNamePrefix;
  } // getFileNamePrefix() 

  // -------------------------------------------------------------------------

  public void setFileNamePrefix(String fileNamePrefix)
  {
    this.fileNamePrefix = fileNamePrefix;
  } // setFileNamePrefix() 

  // -------------------------------------------------------------------------

  public String getFileNameSuffix()
  {
    return fileNameSuffix;
  } // getFileNameSuffix() 

  // -------------------------------------------------------------------------

  public void setFileNameSuffix(String fileNameSuffix)
  {
    this.fileNameSuffix = fileNameSuffix;
  } // setFileNameSuffix() 

  // -------------------------------------------------------------------------

  /**
   * Returns an optional limit for files that are created. A null, zero or
   * negative value mean "no limit".
   */
  public Long getFileSizeLimit()
  {
    return this.fileSizeLimit;
  } // getFileSizeLimit() 

  // -------------------------------------------------------------------------

  /**
   * Sets a limit for files that are created. A null, zero or
   * negative value mean "no limit".
   */
  public void setFileSizeLimit(Long limit)
  {
    this.fileSizeLimit = limit;
  } // setFileSizeLimit() 

  // -------------------------------------------------------------------------

  /**
   * Creates a temporary file and stores the whole contents of the given stream
   * into it.
   * A defined file size limit will be respected and exceeding it will throw an IOException
   * (in which case the target file be deleted).
   * 
   * @param inStream The stream with the data to be stored into the file (must not be null).
   * @return The file descriptor of the created temporary file.
   */
  public File storeStream(InputStream inStream) throws IOException
  {
    File tempFile;

    tempFile = this.createTempFileDescriptor();
    this.storeStreamToFile(inStream, tempFile);
    return tempFile;
  } // storeStream() 

  // -------------------------------------------------------------------------

  /**
   * Creates a temporary file and stores the whole contents of the given stream
   * into it.
   * A defined file size limit will be respected and exceeding it will throw an IOException
   * (inch which case the target file be deleted).
   * 
   * @param inStream The stream with the data to be stored into the file (must not be null).
   * @return The file descriptor of the created temporary file.
   */
  public File storeStream(InputStream inStream, String filenameSnippet) throws IOException
  {
    File tempFile;

    tempFile = this.createTempFileDescriptor(filenameSnippet);
    this.storeStreamToFile(inStream, tempFile);
    return tempFile;
  } // storeStream() 

  // -------------------------------------------------------------------------

  /**
   * Deletes all temporary files in the underlying temp files folder that are older
   * that the given point in time. 
   * 
   * @param timestamp Defines the point in time to compare with.
   * @return The number of deleted files.
   */
  public long removeFilesOlderThan(long timestamp)
  {
    IFileProcessor tempFileKiller;
    ExtendedFileFilter filenameFilter;
    FileWalker fileWalker;
    StringFilterCollection filter;

    filter = new StringFilterCollection(BooleanOperation.AND);
    filter.addFilters(StringPattern.create(TIMESTAMP_SEARCH_PATTERN));
    filenameFilter = new ExtendedFileFilter(filter);
    
    tempFileKiller = new FileDeleteProcessor();
    fileWalker = new FileWalker(tempFileKiller);
    return fileWalker.walkThrough(this.getTempFolderName(), filenameFilter, false);
  } // removeFilesOlderThan()

  // -------------------------------------------------------------------------

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  /**
   * Creates a file for the given file description and stores the whole contents 
   * of the given stream into it.
   * A defined file size limit will be respected and exceeding it will throw an IOException
   * (inch which case the target file be deleted).
   * 
   * @param inStream The stream with the data to be stored into the file (must not be null).
   * @param file The file to which the data of the stream must be written (must not be null).
   */
  protected void storeStreamToFile(InputStream inStream, File file) throws IOException
  {
    try
    {
      FU.createFile(inStream, file, DEFAULT_BUFFER_SIZE, this.getFileSizeLimit());
    }
    catch (IOException ex)
    {
      file.delete();
      throw ex;
    }
  } // storeStreamToFile() 

  // -------------------------------------------------------------------------

  protected void checkTempFolder() throws IOException
  {
    File tempFile;

    // Ensure that it exists
    this.getTempFolder().mkdirs();
    if (!this.getTempFolder().exists())
    {
      throw new FileNotFoundException("Temp folder " + this.getTempFolderName());
    }
    if (!this.getTempFolder().isDirectory())
    {
      throw new IOException("Temp folder " + this.getTempFolderName() + " is no directory!");
    }
    // Now create a temp file in that folder to check if write access is possible.
    tempFile = this.createTempFileDescriptor();
    FU.createFile(tempFile, tempFile.getName().getBytes());
    tempFile.delete();
  } // checkTempFolder() 

  // -------------------------------------------------------------------------

  protected File createTempFileDescriptor()
  {
    return this.createTempFileDescriptor(null);
  } // createTempFileDescriptor() 

  // -------------------------------------------------------------------------

  protected File createTempFileDescriptor(String filenameSnippet)
  {
    File tempFile;
    String fileName;

    fileName = this.createTempFileName(filenameSnippet);
    tempFile = new File(this.getTempFolder(), fileName);
    return tempFile;
  } // createTempFileDescriptor() 

  // -------------------------------------------------------------------------

  protected String createTempFileName(String nameSnippet)
  {
    int id;
    String name = "";

    id = this.getNextNumericId().getAndIncrement();
    if (this.getFileNamePrefix() != null)
    {
      name = this.getFileNamePrefix();
    }
    name = name + NAME_PART_SEPARATOR + this.getTimestamp();
    if (nameSnippet != null)
    {
      name = name + NAME_PART_SEPARATOR + nameSnippet;
    }
    if (this.requiresInstanceUniqueness())
    {
      name = name + NAME_PART_SEPARATOR + this.getUniqueId();
    }
    name = name + NAME_PART_SEPARATOR + SU.leftPad(id, 10);
    if (this.getFileNameSuffix() != null)
    {
      name = name + NAME_PART_SEPARATOR + this.getFileNameSuffix();
    }
    return name;
  } // createTempFileName() 

  // -------------------------------------------------------------------------

  protected String calculateInstanceUniqueId()
  {
    UUID uuid;

    uuid = UUID.randomUUID();
    return uuid.toString();
  } // calculateInstanceUniqueId()

  // -------------------------------------------------------------------------

  protected boolean requiresInstanceUniqueness()
  {
    return this.getUniqueId() != null;
  } // requiresInstanceUniqueness() 

  // -------------------------------------------------------------------------

  protected String getTimestamp()
  {
    SimpleDateFormat formatter;

    formatter = new SimpleDateFormat(TIMESTAMP_FORMAT);
    return formatter.format(new Date());
  } // getTimestamp() 

  // -------------------------------------------------------------------------

  /**
   * Returns the absolute path of the temp folder with forward slash notation. 
   */
  protected String getTempFolderName()
  {
    return FU.standardize(this.getTempFolder().getAbsolutePath());
  } // getTempFolderName() 

  // -------------------------------------------------------------------------

  protected File getTempFolder()
  {
    return this.tempFolder;
  } // getTempFolder() 

  // -------------------------------------------------------------------------

  protected AtomicInteger getNextNumericId()
  {
    return this.nextNumericId;
  } // getNextNumericId() 

  // -------------------------------------------------------------------------

  protected String getUniqueId()
  {
    return this.uniqueId;
  } // getUniqueId() 

  // -------------------------------------------------------------------------

} // class TempFilesManager 
